"""
CBMPy: CBReadtxt module
=======================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBReadtxt.py 710 2020-04-27 14:22:34Z bgoli $)

"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals


from .CBConfig import __CBCONFIG__ as __CBCONFIG__

__DEBUG__ = __CBCONFIG__['DEBUG']
__version__ = __CBCONFIG__['VERSION']

import os
import time
import re
import numpy
from . import CBModel


# userspace
SYMB_SPLIT = ';'
COMPARTMENTS = ['[c]', '[e]', '[p]']
REPLACE_COMPARTMENTS = ['_c', '_e', '_p']
SYMB_IRR = '-->'
SYMB_REV = '<==>'
BIG_BOUND = 999999.0
# userspace


def getBounds(fname, reaction_prefix='R_', has_header=False):
    bndsF = file(fname, 'r')
    bndsLst = []
    bndsOut = {}
    for L in bndsF:
        L = [e.strip() for e in L.split(SYMB_SPLIT)]
        # general replacements '\"' --> '' and '(e)' --> '_e'
        L = [e.replace('\"', '').replace('(e)', '__e') for e in L]
        # print L
        bndsLst.append(L)
    if has_header:
        print(bndsLst.pop(0))
    for b in bndsLst:
        # note the '-' --> '_' and '.' --> '_' replacement in the id
        bndsOut.update(
            {
                reaction_prefix
                + b[0]
                .replace('-', '_')
                .replace('.', '_'): {
                    'min': float(b[1].replace(',', '.')),
                    'max': float(b[2].replace(',', '.')),
                }
            }
        )
    bndsF.close()
    return bndsOut


def getReactions(
    fname,
    reaction_prefix='R_',
    has_header=False,
    save_rpt=False,
    ignore_duplicates=False,
):
    assert os.path.exists(fname), '\n%s is not a valid file!'
    rF = file(fname, 'r')
    rxnLst = []
    Reac = {}
    dupcntr = {}
    real_duplicates = []
    gene_ors = []
    for L in rF:
        L = [e.strip() for e in L.split(SYMB_SPLIT)]
        L = [e.replace('\"', '').replace('(e)', '__e') for e in L]
        rxnLst.append(L)
    if has_header:
        print('Deleting header row:', rxnLst.pop(0))
        time.sleep(2)
    for r in rxnLst:
        assert r[0] != '', '\nReaction with no name: %s' % r
        # print r
        # note the '-' --> '_' and '.' --> '_' replacement in the id
        if len(r) >= 4:
            gene = r[3]
        else:
            gene = 'unknown'
        R_id = reaction_prefix + r[0].replace('-', '_').replace('.', '_')
        ADD_REACTION = True
        if R_id in tuple(Reac):
            if (
                r[1] == Reac[R_id]['eqn']
                and r[2] == Reac[R_id]['name']
                and gene == Reac[R_id]['gene']
            ):
                print('Complete duplicate skipping reaction', R_id)
                if R_id not in real_duplicates:
                    real_duplicates.append(R_id)
                ADD_REACTION = False
            # elif r[1] == Reac[R_id]['eqn'] and r[2] == Reac[R_id]['name']:
            elif r[1] == Reac[R_id]['eqn']:
                Reac[R_id]['gene'] = Reac[R_id]['gene'] + ' OR %s' % gene.strip()
                print(
                    'Reaction differs only by gene association adding OR association',
                    R_id,
                    Reac[R_id]['gene'],
                )
                ADD_REACTION = False
                if R_id not in gene_ors:
                    gene_ors.append(R_id)
                if r[2] not in Reac[R_id]['name']:
                    Reac[R_id]['name'] = '%s OR %s' % (Reac[R_id]['name'], r[2])
            else:
                if ignore_duplicates:
                    ADD_REACTION = False
                    print('Ignoring duplicate reaction creating new reaction:', R_id)
                else:
                    if R_id in dupcntr:
                        dupcntr[R_id] = dupcntr[R_id] + 1
                    else:
                        dupcntr.update({R_id: 1})
                    R_id = '%s_%i' % (R_id, dupcntr[R_id])
                    ADD_REACTION = True
                    print('Duplicate reaction creating new reaction:', R_id)

        if ADD_REACTION:
            ##  Reac.update({R_id : {'id' : R_id, 'name' : r[1], 'eqn' : r[2], 'gene' : gene} })
            Reac.update({R_id: {'id': R_id, 'name': r[2], 'eqn': r[1], 'gene': gene}})
    rF.close()

    if save_rpt:
        F = open('csv_read_rpt.txt', 'w')
        F.write('\nCSV Read Report\n********************\n')
        F.write('\nDuplicate entries ignored:\n')
        for d in real_duplicates:
            F.write('\t%s\n' % d)
        F.write('\nDuplicate reactions renamed (instances):\n')
        for d in dupcntr:
            F.write('\t%s : %i\n' % (d, dupcntr[d] + 1))
        if len(dupcntr) < 1:
            F.write('\tNone\n')
        F.write('\nDuplicate reactions used for gene association:\n')
        for d in gene_ors:
            F.write('\n\t%s\n\t%s\n\t%s\n' % (d, Reac[d]['gene'], Reac[d]['name']))
        F.close()

    return Reac


def getReactions_old_format(
    fname,
    reaction_prefix='R_',
    has_header=False,
    save_rpt=False,
    ignore_duplicates=False,
):
    assert os.path.exists(fname), '\n%s is not a valid file!'
    rF = file(fname, 'r')
    rxnLst = []
    Reac = {}
    dupcntr = {}
    real_duplicates = []
    gene_ors = []
    for L in rF:
        L = [e.strip() for e in L.split(SYMB_SPLIT)]
        L = [e.replace('\"', '').replace('(e)', '__e') for e in L]
        rxnLst.append(L)
    if has_header:
        print('Deleting header row:', rxnLst.pop(0))
        time.sleep(2)
    for r in rxnLst:
        assert r[0] != '', '\nReaction with no name: %s' % r
        # print r
        # note the '-' --> '_' and '.' --> '_' replacement in the id
        if len(r) >= 4:
            gene = r[3]
        else:
            gene = 'unknown'
        R_id = reaction_prefix + r[0].replace('-', '_').replace('.', '_')
        ADD_REACTION = True
        if R_id in tuple(Reac):
            if (
                r[1] == Reac[R_id]['eqn']
                and r[2] == Reac[R_id]['name']
                and gene == Reac[R_id]['gene']
            ):
                print('Complete duplicate skipping reaction', R_id)
                if R_id not in real_duplicates:
                    real_duplicates.append(R_id)
                ADD_REACTION = False
            # elif r[1] == Reac[R_id]['eqn'] and r[2] == Reac[R_id]['name']:
            elif r[1] == Reac[R_id]['eqn']:
                Reac[R_id]['gene'] = Reac[R_id]['gene'] + ' OR %s' % gene.strip()
                print(
                    'Reaction differs only by gene association adding OR association',
                    R_id,
                    Reac[R_id]['gene'],
                )
                ADD_REACTION = False
                if R_id not in gene_ors:
                    gene_ors.append(R_id)
                if r[2] not in Reac[R_id]['name']:
                    Reac[R_id]['name'] = '%s OR %s' % (Reac[R_id]['name'], r[2])
            else:
                if ignore_duplicates:
                    ADD_REACTION = False
                    print('Trying to create duplicate reaction, ignoring:', R_id)
                else:
                    if R_id in dupcntr:
                        dupcntr[R_id] = dupcntr[R_id] + 1
                    else:
                        dupcntr.update({R_id: 1})
                    R_id = '%s_%i' % (R_id, dupcntr[R_id])
                    ADD_REACTION = True
                    print('Duplicate reaction, creating new reaction:', R_id)

        if ADD_REACTION:
            ##  Reac.update({R_id : {'id' : R_id, 'name' : r[1], 'eqn' : r[2], 'gene' : gene} })
            Reac.update({R_id: {'id': R_id, 'name': r[1], 'eqn': r[2], 'gene': gene}})
    rF.close()

    if save_rpt:
        F = open('csv_read_rpt.txt', 'w')
        F.write('\nCSV Read Report\n********************\n')
        F.write('\nDuplicate entries ignored:\n')
        for d in real_duplicates:
            F.write('\t%s\n' % d)
        F.write('\nDuplicate reactions renamed (instances):\n')
        for d in dupcntr:
            F.write('\t%s : %i\n' % (d, dupcntr[d] + 1))
        if len(dupcntr) < 1:
            F.write('\tNone\n')
        F.write('\nDuplicate reactions used for gene association:\n')
        for d in gene_ors:
            F.write('\n\t%s\n\t%s\n\t%s\n' % (d, Reac[d]['gene'], Reac[d]['name']))
        F.close()

    return Reac


def parseReactions(sRxns):
    Reactions = {}
    for RXN in sRxns:
        R = sRxns[RXN]['eqn']
        Rcmp = None
        Reversible = None
        Exchange = False
        if sRxns[RXN]['id'][-3:] == '__e':
            Exchange = True
        if SYMB_REV in R:
            Reversible = True
            LHS, RHS = R.split(SYMB_REV)
        elif SYMB_IRR in R:
            Reversible = False
            LHS, RHS = R.split(SYMB_IRR)
        else:
            print('No reversability information in reaction: %s' % RXN)
            print(R)
            raw_input('\nPress any key to continue ...\n')
        if LHS[:3] in COMPARTMENTS:
            Rcmp = R[:3]
            cmp, LHS = [i.strip() for i in LHS.split(':')]

            if len(LHS) > 0:
                LHS = [i.strip().replace('-', '_') for i in LHS.split('+')]
                LHS = [i + REPLACE_COMPARTMENTS[COMPARTMENTS.index(cmp)] for i in LHS]
            if len(RHS) > 0:
                RHS = [i.strip().replace('-', '_') for i in RHS.split('+')]
                RHS = [i + REPLACE_COMPARTMENTS[COMPARTMENTS.index(cmp)] for i in RHS]
        else:
            LHS = [r.strip().replace('-', '_') for r in LHS.split('+')]
            RHS = [r.strip().replace('-', '_') for r in RHS.split('+')]
            for c in range(len(COMPARTMENTS)):
                LHS = [r.replace(COMPARTMENTS[c], REPLACE_COMPARTMENTS[c]) for r in LHS]
                RHS = [r.replace(COMPARTMENTS[c], REPLACE_COMPARTMENTS[c]) for r in RHS]

        cf = re.compile("\(.*\)")
        LHS = [[re.findall(cf, l), l] for l in LHS]
        RHS = [[re.findall(cf, r), r] for r in RHS]

        for l in range(len(LHS)):
            if LHS[l][1] == '_c':
                raw_input('Ooops1', LHS)
            if len(LHS[l][0]) == 0:
                LHS[l][0] = 1.0
                LHS[l][1] = 'M_' + LHS[l][1].replace('.', '_')
            else:
                LHS[l][1] = 'M_' + LHS[l][1].replace(LHS[l][0][0], '').strip().replace(
                    '.', '_'
                )
                LHS[l][0] = float(LHS[l][0][0].replace('(', '').replace(')', ''))
            LHS[l] = tuple(LHS[l])

        for r in range(len(RHS)):
            if RHS[r][1] == '_c':
                raw_input('Ooops2', RHS)
            if len(RHS[r][0]) == 0:
                RHS[r][0] = 1.0
                RHS[r][1] = 'M_' + RHS[r][1].replace('.', '_')
            else:
                RHS[r][1] = 'M_' + RHS[r][1].replace(RHS[r][0][0], '').strip().replace(
                    '.', '_'
                )
                RHS[r][0] = float(RHS[r][0][0].replace('(', '').replace(')', ''))
            RHS[r] = tuple(RHS[r])
        geneAss = 'unknown'
        if 'gene' in sRxns[RXN]:
            geneAss = sRxns[RXN]['gene']
        Reactions.update(
            {
                RXN: {
                    'id': RXN,
                    'reversible': Reversible,
                    'exchange': Exchange,
                    'substrates': LHS,
                    'products': RHS,
                    'name': sRxns[RXN]['name'],
                    'gene': geneAss,
                }
            }
        )
    return Reactions


def addBoundsToReactions(Reactions, Bounds, default=1111.0):
    default = abs(default)
    for r in Reactions:
        if r in Bounds:
            Reactions[r].update({'min': Bounds[r]['min'], 'max': Bounds[r]['max']})
        else:
            # print '\t:'
            if Reactions[r]['reversible']:
                Reactions[r].update({'min': -default, 'max': default})
            else:
                Reactions[r].update({'min': 0.0, 'max': default})


def dumpReactionsToTxt(Reactions, fname):
    dF = file(fname, 'w')
    for rx in Reactions:
        dF.write('\n%s\n' % Reactions[rx]['id'])
        dF.write('Name: %s\n' % Reactions[rx]['name'])
        dF.write('Reversible: %s\n' % Reactions[rx]['reversible'])
        dF.write('Exchange: %s\n' % Reactions[rx]['exchange'])
        dF.write(
            'Bounds: min=%s max=%s\n' % (Reactions[rx]['min'], Reactions[rx]['max'])
        )
        if Reactions[rx]['reversible']:
            rsign = '<==>'
        else:
            rsign = '-->'

        st = ''
        for s in Reactions[rx]['substrates']:
            st += '%s %s + ' % s
        st = st[:-2]
        pt = ''
        for p in Reactions[rx]['products']:
            pt += '%s %s + ' % p
        pt = pt[:-2]
        dF.write('%s %s %s\n' % (st.strip(), rsign.strip(), pt.strip()))
    dF.flush()
    dF.close()


def getSpecies(reactions):
    species = {}
    for r in reactions:
        for s in reactions[r]['substrates']:
            species.update({s[1]: {'compartment': s[1][-1]}})
        for p in reactions[r]['products']:
            species.update({p[1]: {'compartment': p[1][-1]}})
    return species


def dumpSpeciesToTxt(species, fname):
    dF = file(fname, 'w')
    for s in species:
        dF.write('\n%s\n' % s)
        dF.write('Compartment: %s\n' % species[s]['compartment'])
    dF.flush()
    dF.close()


def buildFBAobj(reactions, species, objfname, modname='model'):
    fm = CBModel
    M = fm.Model(modname)
    M.setName(modname)
    M.description = modname
    for s in species:
        S = fm.Species(s, compartment=species[s]['compartment'])
        S.value = 0.0
        S.setName(s)
        M.addSpecies(S)
    for r in reactions:
        R = fm.Reaction(
            r, name=reactions[r]['name'], reversible=reactions[r]['reversible']
        )
        for su in reactions[r]['substrates']:
            SU = fm.Reagent(r + su[1], su[1], -su[0])
            R.addReagent(SU)
            spx = M.getSpecies(su[1])
            if su[1] not in spx.isReagentOf():
                spx.reagent_of.append(R.getId())
        for pr in reactions[r]['products']:
            PR = fm.Reagent(r + pr[1], pr[1], pr[0])
            R.addReagent(PR)
            spx = M.getSpecies(pr[1])
            if pr[1] not in spx.isReagentOf():
                spx.reagent_of.append(R.getId())
        R.is_exchange = reactions[r]['exchange']
        LB = fm.FluxBound(r + 'min', r, 'greaterEqual', reactions[r]['min'])
        UB = fm.FluxBound(r + 'max', r, 'lessEqual', reactions[r]['max'])
        R.value = 0.0
        if 'gene' in reactions[r]:
            R.setAnnotation('GENE ASSOCIATION', reactions[r]['gene'])
        M.addReaction(R, create_default_bounds=False)
        M.addFluxBound(LB)
        M.addFluxBound(UB)
    if objfname != None:
        OB = fm.Objective('objective1', 'maximize')
        FO = fm.FluxObjective(objfname + '_obj', objfname)
        OB.addFluxObjective(FO)
        M.addObjective(OB, active=True)
    return M


def readCSV(
    model_file,
    bounds_file=None,
    biomass_flux=None,
    model_id='FBAModel',
    reaction_prefix='R_',
    has_header=False,
):
    """
    This function loads a CSV file and translates it into a Python object::

     - *model_file* the name of the CSV file that contains the model
     - *bounds_file* the name of the CSV file that contains the flux bounds
     - *biomass_flux* the name of the reaction that is the objective function
     - *reaction_prefix* [default='R _'] the prefix to add to input reaction ID's
     - *has_header* [default=False] if there is a header row in the csv file

    """

    assert os.path.exists(model_file), '\nPlease supply a file name that exists.'
    Rdict = getReactions(
        model_file,
        reaction_prefix=reaction_prefix,
        has_header=has_header,
        save_rpt=True,
    )
    R = parseReactions(Rdict)
    del Rdict
    if bounds_file != None and os.path.exists(bounds_file):
        B = getBounds(
            bounds_file, reaction_prefix=reaction_prefix, has_header=has_header
        )
        print('\nBounds loaded from file: {}\n'.format(bounds_file))
    else:
        B = {}

    addBoundsToReactions(R, B, default=BIG_BOUND)
    S = getSpecies(R)
    dumpReactionsToTxt(R, 'dump_reactions.txt')
    dumpSpeciesToTxt(S, 'dump_species.txt')

    fba = buildFBAobj(reactions=R, species=S, objfname=biomass_flux, modname=model_id)
    return fba


"""
if __name__ == '__main__':
    cDir = os.path.dirname(os.path.abspath(os.sys.argv[0]))
    work_dir = os.path.join(cDir, 'spy')

    Bounds = getBounds(os.path.join(work_dir, 'Spy_bounds_semicolon.csv'))
    Reactions  = getReactions(os.path.join(work_dir, 'Spy_reaction_biomass_semicolon.csv'))
    ##  Bounds = getBounds(os.path.join(work_dir, 'boundaries_EFA.csv'))
    ##  Reactions  = getReactions(os.path.join(work_dir, 'reactions_EFA1.csv'))

    addBoundsToReactions(Reactions, Bounds, default=1000.0)
    dumpReactionsToTxt(Reactions, 'dump_reactions.txt')
    Species = getSpecies(Reactions)
    dumpSpeciesToTxt(Species, 'dump_species.txt')
    fba = buildFBAobj(Reactions, Species, 'biomass_LPL6_0', 'Spy')
"""



'''
Created on Nov 17, 2014

@author: arne
'''
import unittest

import os
from math import isnan

from cbmpy import CBRead
from cbmpy import CBTools
from cbmpy import CBModel

from cbmpy.solver import CBSolver

from .enumerate import EFMEnumerator
from .sparserationals import Matrix
from .decomposition import Decomposition, Vertex
from . import matroid

class TestEnumerateToy(unittest.TestCase):

    def setUp(self):
        cDir = os.path.dirname(__file__)
        model_dir = os.path.join(cDir, '../models')
        model_file = "toy_model_maarleveld2014.l3.xml"

        self.cmod = CBRead.readSBML3FBC(model_file, model_dir)
        self.cmod = CBTools.splitReversibleReactions(self.cmod)

        r = self.cmod.getReaction('R01')
        print(r.getUpperBound())
        print(r.getLowerBound())
        r.setLowerBound(2.0)

        for r in self.cmod.reactions:
            r.setUpperBound(min(1000, r.getUpperBound()))

        self.cmod.buildStoichMatrix()

        matrix = Matrix()
        labels, _ = matrix.addMetabolicNetwork(self.cmod)
        self.matroid = matroid.fromMatrix(matrix, labels)

        self.decomp = Decomposition(self.matroid)
        sim = self.matroid.similarityMatrix("combined")
        self.decomp.poolmanMethod(self.decomp.root, sim)

    def tearDown(self):
        pass

    def testEnumerate(self):
        CBSolver.FluxVariabilityAnalysis(self.cmod, objF2constr=False)
        enum = EFMEnumerator(self.cmod, self.decomp.root, None)
        F = enum.enumerateMinimal()
        print(F)
        self.assertEquals(F, set(
            [frozenset(['R01', 'R17', 'R05', 'R10', 'R11', 'R09', 'R18', 'R03_fwd', 'R13_fwd', 'R12', 'R16', 'R14_fwd', 'R04_fwd']),
             frozenset(['R01', 'R17', 'R05', 'R10', 'R11', 'R09', 'R18', 'R02_fwd', 'R12', 'R16', 'R14_fwd', 'R13_fwd']),
             frozenset(['R01', 'R17', 'R05', 'R07', 'R11', 'R08', 'R06', 'R18', 'R03_fwd', 'R13_fwd', 'R12', 'R16', 'R14_fwd', 'R04_fwd']),
             frozenset(['R01', 'R17', 'R05', 'R07', 'R11', 'R08', 'R06', 'R18', 'R02_fwd', 'R12', 'R16', 'R14_fwd', 'R13_fwd']),
             frozenset(['R01', 'R17', 'R15', 'R05', 'R10', 'R11', 'R09', 'R18', 'R03_fwd', 'R12', 'R16', 'R04_fwd']),
             frozenset(['R01', 'R17', 'R15', 'R05', 'R10', 'R11', 'R09', 'R18', 'R03_fwd', 'R14_rev', 'R13_rev', 'R16', 'R04_fwd']),
             frozenset(['R01', 'R17', 'R15', 'R05', 'R10', 'R11', 'R09', 'R18', 'R02_fwd', 'R12', 'R16']),
             frozenset(['R01', 'R17', 'R15', 'R05', 'R10', 'R11', 'R09', 'R18', 'R02_fwd', 'R14_rev', 'R16', 'R13_rev']),
             frozenset(['R01', 'R17', 'R15', 'R05', 'R07', 'R11', 'R08', 'R06', 'R18', 'R03_fwd', 'R12', 'R16', 'R04_fwd']),
             frozenset(['R01', 'R17', 'R15', 'R05', 'R07', 'R11', 'R08', 'R06', 'R18', 'R03_fwd', 'R14_rev', 'R13_rev', 'R16', 'R04_fwd']),
             frozenset(['R01', 'R17', 'R15', 'R05', 'R07', 'R11', 'R08', 'R06', 'R18', 'R02_fwd', 'R14_rev', 'R16', 'R13_rev']),
             frozenset(['R01', 'R17', 'R15', 'R05', 'R07', 'R11', 'R08', 'R06', 'R18', 'R02_fwd', 'R12', 'R16'])]
        )
        )
        evaluateReuse(F, self.decomp.root, None)


@unittest.skip("takes quite long")
class TestEnumerateECore(unittest.TestCase):

    def setUp(self):
        cDir = os.path.dirname(__file__)
        model_dir = os.path.join(cDir, '../models')
        model_file = "ecoli_core_COBRA.xml"

        self.cmod = CBRead.readCOBRASBML(model_file, model_dir)
        self.cmod = CBTools.splitReversibleReactions(self.cmod)
        for r in self.cmod.reactions:
            r.setUpperBound(min(1000, r.getUpperBound()))
            r.setLowerBound(0)

        r = self.cmod.getReaction('R_Biomass_Ecoli_core_w_GAM')
        print(r.getUpperBound())
        print(r.getLowerBound())
        r.setLowerBound(0.1)
        r.setUpperBound(0.1)

        self.cmod.buildStoichMatrix()

        matrix = Matrix()
        labels, _ = matrix.addMetabolicNetwork(self.cmod)
        self.matroid = matroid.fromMatrix(matrix, labels)

        self.decomp = Decomposition(self.matroid)
        self.decomp.splitSimple()
        for v in self.decomp.listNonLeaves():
            if not v.isSimple():
                sim = v.matroid.similarityMatrix("combined")
                self.decomp.poolmanMethod(v, sim)

        write(self.decomp.makeGraphViz(cmod=self.cmod), "ecoli_core_split_poolman_combined.gv")
        # self.assertTrue(self.decomp.verifyFullyBranched())
        # self.assertTrue(self.decomp.verifyEdgeWidth())
        print("width = %d" % self.decomp.getWidth())

    def tearDown(self):
        pass

    def testEnumerate(self):
        CBSolver.FluxVariabilityAnalysis(self.cmod, objF2constr=False)
        enum = EFMEnumerator(self.cmod, self.decomp.root, None)
        fObj = CBModel.FluxObjective('R_EX_glc_e_rev_obj', reaction='R_EX_glc_e_rev', coefficient=1)
        objF = CBModel.Objective('minimize_nutrient', 'min')
        objF.addFluxObjective(fObj)
        self.cmod.addObjective(objF, active=True)
        CBSolver.analyzeModel(self.cmod)
        #CBSolver.MinimizeSumOfAbsFluxes(self.cmod, objF2constr=False)
        face = set()
        for r in self.cmod.reactions:
            if r.getValue() > 1e-5:
                face.add(r.getId())
                print('%s: %g' % (r.getId(), r.getValue()))
        self.assertTrue(enum.test(face))
        F = enum.enumerateMinimal()
        print(F)


def evaluateReuse(F, node, parent):
    mod = node.getLeaves(parent)
    Fmod = {}
    for f in F:
        fmod = f.intersection(mod)
        Fmod[fmod] = Fmod.get(fmod, 0) + 1
    FmodVals = Fmod.values()
    FmodVals.sort()
    FmodVals.reverse()
    print(mod)
    print(FmodVals)
    if len(mod) > 10 and isinstance(node, Vertex):
        adj = node.getAdjacent()
        for a in adj:
            if a != parent:
                evaluateReuse(F, a, node)


@unittest.skip("takes quite long")
class TestEnumerateEcoliTextbookBigg(unittest.TestCase):

    def setUp(self):
        cDir = os.path.dirname(__file__)
        model_dir = os.path.join(cDir, '../models')
        model_file = "bigg_textbook.xml.l3fbc.xml"

        self.cmod = CBRead.readSBML3FBC(model_file, model_dir)
        self.cmod = CBTools.splitReversibleReactions(self.cmod)
        for r in self.cmod.reactions:
            if r.getUpperBound() != r.getLowerBound():
                r.setUpperBound(min(1000, r.getUpperBound()))
                r.setLowerBound(0)

        r = self.cmod.getReaction('R_Biomass_Ecoli_core_N_LPAREN_w_FSLASH_GAM_RPAREN__Nmet2')
        print(r.getUpperBound())
        print(r.getLowerBound())
        r.setLowerBound(0.8)
        r.setUpperBound(0.8)

        self.cmod.buildStoichMatrix()

        matrix = Matrix()
        labels, _ = matrix.addMetabolicNetwork(self.cmod)
        self.matroid = matroid.fromMatrix(matrix, labels)

        self.decomp = Decomposition(self.matroid)
        self.decomp.splitSimple()
        for v in self.decomp.listNonLeaves():
            if not v.isSimple():
                sim = v.matroid.similarityMatrix("combined")
                self.decomp.poolmanMethod(v, sim)

        write(self.decomp.makeGraphViz(cmod=self.cmod), "ecoli_textbook_bigg_split_poolman_combined.gv")
        # self.assertTrue(self.decomp.verifyFullyBranched())
        # self.assertTrue(self.decomp.verifyEdgeWidth())
        print("width = %d" % self.decomp.getWidth())

    def tearDown(self):
        pass

    def testEnumerateSubnetwork(self):
        CBSolver.FluxVariabilityAnalysis(self.cmod, objF2constr=False)
        v = self.decomp.getVertex(247)
        exclude = self.decomp.getVertex(283)
        enum = EFMEnumerator(self.cmod, v, exclude)
        F = enum.enumerateMinimal()
        print(F)
        lps = self.buildStrictLP(enum)
        #lpw = self.buildWeakLP(enum)
        modmatroid = self.matroid.contract([r for r in self.matroid.elems if r not in enum.module and r not in enum.fixed])
        # print(modmatroid.basis)
        # print(modmatroid.nbasis)
        print(enum.fixed)
        res = []
        for f in F:
            if self.isVertex(modmatroid, f):
                if not enum.isFeasible(f):
                    assert False
                sol = enum.lp.getSolution()
                interface = None
                for r in enum.module:
                    col = self.cmod.N.getColsByName(r)
                    if interface == None:
                        interface = col * sol[r]
                    else:
                        interface += col * sol[r]

                # check strict feasibility
                res.append((interface, self.isFeasible(enum, lps, f)))
                print(f)
                for m in range(len(interface)):
                    v = interface[m]
                    if abs(v) > 1e-4:
                        print('%s: %g' % (self.cmod.N.row[m], v))
        # print(res)

    def isVertex(self, modmatroid, face):
        return modmatroid.isIndependent(face)

    def isFeasible(self, enum, lp, face):
        bounds = {}
        for r in enum.module:
            if r in face:
                # set lower bound to self.minflux
                lb = max(1e-4, self.cmod.getReactionLowerBound(r))
                ub = self.cmod.getReactionUpperBound(r)
                if lb > ub:
                    return False
                bounds[r] = (lb, ub)
            else:
                # fix flux to zero
                lb = self.cmod.getReactionLowerBound(r)
                if lb > 0:
                    return False
                bounds[r] = (lb, 0)

        lp.setBounds(bounds)
        lp.solve()

        if lp.getSolutionStatus() != 'LPS_UNDEF':
            return lp.isFeasible()
        else:
            print('warning: unable to solve LP')
            return True  # if we keep it, we don't miss solutions

    def buildWeakLP(self, enum):
        lp = CBSolver.createSolver(self.cmod)
        for r in self.mnet.reactions:
            if r.getId() not in self.module:
                # remove all bounds from the variable
                # unless it has a fixed flux rate
                span = r.getFVAdata()[3]
                if (span > self.tol or isnan(span)):
                    # remove flux bounds
                    lp.setLowerBounds({r.getId(): None})
                    lp.setUpperBounds({r.getId(): None})
        # clear objective function
        lp.setObjective(reset=True)
        return lp

    def buildStrictLP(self, enum):
        lp = CBSolver.createSolver(self.cmod)
        # clear objective function
        lp.setObjective(reset=True)
        return lp

    @unittest.skip("takes quite long")
    def testEnumerate(self):
        CBSolver.FluxVariabilityAnalysis(self.cmod, objF2constr=False)
        enum = EFMEnumerator(self.cmod, self.decomp.root, None)
        fObj = CBModel.FluxObjective('R_EX_glc_e_rev_obj', reaction='R_EX_glc_LPAREN_e_RPAREN__rev', coefficient=1)
        objF = CBModel.Objective('minimize_nutrient', 'min')
        objF.addFluxObjective(fObj)
        self.cmod.addObjective(objF, active=True)
        CBSolver.analyzeModel(self.cmod)
        #CBSolver.MinimizeSumOfAbsFluxes(self.cmod, objF2constr=False)
        face = set()
        for r in self.cmod.reactions:
            if r.getValue() > 1e-5:
                face.add(r.getId())
                print('%s: %g' % (r.getId(), r.getValue()))
        self.assertTrue(enum.test(face))
        F = enum.enumerateMinimal()
        print(F)


def write(result, filename):
    cDir = os.path.dirname(__file__)
    result_dir = os.path.join(cDir, '../../results')
    result_file = os.path.join(result_dir, filename)
    fo = open(result_file, mode='w')
    fo.write(result)
    fo.close()


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
CBMPy: CBSandBox module
=======================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBSandBox.py 710 2020-04-27 14:22:34Z bgoli $)

"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals

import os
import time
import re

try:
    import pickle
except ImportError:
    import cPickle as pickle
import numpy


from .CBConfig import __CBCONFIG__ as __CBCONFIG__

__DEBUG__ = __CBCONFIG__['DEBUG']
__version__ = __CBCONFIG__['VERSION']

# Set Suffix new function extended by Ruchir


def setSuffix_extended(model, suffix='', target='all', ignore_list=[]):
    print(
        '\nThis function has been replaced by the new model method: cmod.renameObjectIds()\nPlease update your code!\n'
    )
    t = 6
    while t > 0:
        print(t)
        time.sleep(1)
        t = t - 1
    if hasattr(model, 'N') and hasattr(model.N, 'col'):
        assert len(model.N.col) == len(
            model.reactions
        ), "\nStoichiometry (number of reactions) doesn't match the model"
    if hasattr(model, 'N') and hasattr(model.N, 'row'):
        assert len(model.N.row) == len(
            model.species
        ), "\nStoichiometry (number of species) doesn't match the model"

    assert type(suffix) == str, "\nSuffix must be a string not %s" % suffix
    assert target in ['species', 'reactions', 'all'], (
        "\nTarget can only be one of ['species', 'reactions','all'] not %s" % target
    )

    model.suffix = suffix

    if target == 'species' or target == 'all':
        for s in model.species:
            if s.getId() not in ignore_list:
                s.id = s.id + suffix
    if hasattr(model, 'N'):
        if hasattr(model.N, 'row'):
            rowidx = model.N.row.index(s)
            model.N.row[rowidx] = model.N.row[rowidx] + suffix

    for r in range(len(s.isReagentOf())):
        if s.reagent_of[r] not in ignore_list:
            s.reagent_of[r] = s.reagent_of[r] + suffix

    if target == 'reactions' or target == 'all':
        for r in model.reactions:
            if r.getId() not in ignore_list:
                r.id = r.id + suffix
            for reag in r.reagents:
                if reag.species_ref not in ignore_list:
                    reag.species_ref = reag.species_ref + suffix
            if hasattr(model, 'N'):
                if hasattr(model.N, 'col'):
                    colidx = model.N.col.index(r)
                    model.N.col[colidx] = model.N.col[colidx] + suffix
    if target == 'constraints' or target == 'reactions' or target == 'all':
        for c in model.flux_bounds:
            if c.reaction not in ignore_list:
                c.id = c.id + suffix
                c.reaction = c.reaction + suffix

    if target == 'objectives' or target == 'all':
        for obj in model.objectives:
            obj.id = obj.id + suffix
        for f in obj.fluxObjectives:
            if f.reaction not in ignore_list:
                f.id = f.id + suffix
                f.reaction = f.reaction + suffix



"""
CBMPy: __init__.py
==================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: __init__.py 710 2020-04-27 14:22:34Z bgoli $)

"""
#
# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals

import os

# Setup some environment
__SILENT_START__ = False
from . import CBConfig

__CBCONFIG__ = CBConfig.__CBCONFIG__
__version__ = CBConfig.__CBCONFIG__['VERSION']
__CBCONFIG__['CBMPY_DIR'] = os.path.split(CBConfig.__file__)[0]

# This is just a hack for backwards compatibility with existing scripts
# that enables "from cbmpy import *" to work - bgoli

# lets add some shortcuts that will be core commands
__HAVE_SOLVER__ = False
try:
    from . import CBSolver

    analyzeModel = CBSolver.analyzeModel
    FluxVariabilityAnalysis = CBSolver.FluxVariabilityAnalysis
    MinimizeSumOfAbsFluxes = CBSolver.MinimizeSumOfAbsFluxes
    doFBA = CBSolver.analyzeModel
    doFVA = CBSolver.FluxVariabilityAnalysis
    doFBAMinSum = CBSolver.MinimizeSumOfAbsFluxes
    __HAVE_SOLVER__ = True
except (ImportError, AttributeError):
    print('No solver present, unable to create shortcuts')

from .CBWrite import writeFVAtoCSV, writeModelToExcel97
from .CBXML import _HAVE_SBML_

if _HAVE_SBML_:
    from .CBXML import sbml_fileFindVersion, sbml_fileValidate
    from .CBRead import loadModel, readSBML3FBC, readSBML2FBA, readCOBRASBML
    from .CBWrite import (
        saveModel,
        writeSBML3FBC,
        writeModelToExcel97,
        writeModelToCOMBINEarchive,
        writeCOBRASBML,
        writeSBML3FBCV2,
    )
else:
    print('\nWARNING: No SBML support, top-level SBML read/write functions disabled.\n')

try:
    from . import CBMultiCore
except ImportError as ex:
    print(ex)
    print('MultiCore module not available')
    CBMultiCore = None

INF = float('inf')
NINF = -float('inf')
NAN = float('nan')

from . import (
    CBModel,
    CBCommon,
    CBModelTools,
    CBRead,
    CBReadtxt,
    CBTools,
    CBWrite,
    CBXML,
    CBNetDB,
    CBPlot,
    CBMultiModel,
)
from .CBTools import deSerialize, deSerializeFromDisk
from . import PyscesSED

SED = PyscesSED.SEDCBMPY


def test():
    print(
        '\nCBMPy test suite is now separate from CMBPy source, see https://github.com/SystemsBioinformatics/cbmpy for details.\n'
    )


del os

if not __SILENT_START__:
    print('\n\n***********************************************************************')
    print(
        '* Welcome to CBMPy ({}) - PySCeS Constraint Based Modelling        *'.format(
            CBConfig.__CBCONFIG__['VERSION']
        )
    )
    print('*                http://cbmpy.sourceforge.net                         *')
    print('* Copyright(C) Brett G. Olivier 2014 - 2020                           *')
    print('* Systems Biology Lab, Vrije Universiteit Amsterdam                   *')
    print('* Amsterdam, The Netherlands                                          *')
    print('* CBMPy is developed as part of the BeBasic MetaToolKit Project       *')
    print('* Distributed under the GNU GPL v 3.0 licence, see                    *')
    print('* LICENCE (supplied with this release) for details                    *')
    print('***********************************************************************\n')
